import pandas as pd
from pathlib import Path
from sklearn.model_selection import train_test_split


class BaseDataset:

    def __init__(
            self,
            train_data_path: str,
            test_data_path: str = None,
            random_state: int = 42,
            text_label: str = '',
            target_label: str = '',
            test_size: float = 0.3,
    ):
        self.train = self.read_data(Path(repr(train_data_path)[1:-1]).as_posix())[[text_label, target_label]]
        self.train.dropna(inplace=True)
        self.test = self.read_data(
            Path(repr(test_data_path)[1:-1]).as_posix()
        )[[text_label, target_label]].dropna() if test_data_path is not None else None

        if not self.test:
            self.__train_test_split(
                random_state=random_state,
                target_label=target_label,
                test_size=test_size
            )

    def read_data(self, path: str) -> pd.DataFrame:
        raise NotImplementedError

    def __train_test_split(self, random_state: int, target_label: str, test_size: float):
        self.train, self.test = train_test_split(self.train,
                                                 test_size=test_size,
                                                 random_state=random_state,
                                                 stratify=self.train[target_label]
                                                 )



import torch
from sklearn.metrics import classification_report
from bert_clf.src.BertCLF import BertCLF
from bert_clf.src.early_stopping import EarlyStopping
import numpy as np
from sklearn.metrics import f1_score
from tqdm import tqdm
from typing import Dict, Any


def predict_metrics(
        model: BertCLF,
        iterator: torch.utils.data.DataLoader,
):

    """
    :param model: model architecture that you want to fine-tune
    :param iterator: iterator with data reserved for evaluating
    """

    true = []
    pred = []

    model.eval()
    with torch.no_grad():
        for texts, ys in tqdm(iterator, total=len(iterator), desc="Computing final metrics..."):
            predictions = model(texts.to(model.device)).squeeze()
            preds = predictions.detach().to('cpu').numpy().argmax(1).tolist()
            y_true = ys.tolist()
            true.extend(y_true)
            pred.extend(preds)

    true = [model.mapper[x] for x in true]
    pred = [model.mapper[x] for x in pred]

    print(classification_report(true, pred, zero_division=0))


def train(
        model: BertCLF,
        iterator: torch.utils.data.DataLoader,
        optimizer: torch.optim,
        criterion: torch.nn,
        average: str = 'macro'
) -> float:
    """
    :param model: model architecture that you want to fine-tune
    :param iterator: iterator with data reserved for bert_clf
    :param optimizer: torch optimizer
    :param criterion: instance of torch-like loses
    :param average: type of averaging for f1 sklearn metric. Possible types are: 'micro', 'macro', 'weighted'
    :return: mean metric for the bert_clf loop
    """

    epoch_loss = []
    epoch_f1 = []

    model.train()

    for texts, ys in tqdm(iterator, total=len(iterator), desc='Training loop'):

        optimizer.zero_grad()
        predictions = model(texts.to(model.device))
        loss = criterion(predictions, ys.to(model.device))

        loss.backward()
        optimizer.step()
        preds = predictions.detach().to('cpu').numpy().argmax(1).tolist()
        y_true = ys.tolist()

        epoch_loss.append(loss.item())
        epoch_f1.append(f1_score(y_true, preds, average=average))

    return np.mean(epoch_f1)


def evaluate(
        model: BertCLF,
        iterator: torch.utils.data.DataLoader,
        criterion: torch.nn,
        average: str = 'macro'
) -> float:
    """
    :param model: trained model (instance of class model.CLF)
    :param iterator: instance of torch.utils.data.DataLoader
    :param criterion: instance of torch-like loses
    :param average: type of averaging for f1 sklearn metric. Possible types are: 'micro', 'macro', 'weighted'
    :return: mean metric for the evaluating loop
    """
    epoch_loss = []
    epoch_f1 = []

    model.eval()
    with torch.no_grad():
        for texts, ys in tqdm(iterator, total=len(iterator), desc='Evaluating loop'):
            predictions = model(texts.to(model.device))
            loss = criterion(predictions, ys.to(model.device))
            preds = predictions.detach().to('cpu').numpy().argmax(1).tolist()
            y_true = ys.tolist()

            epoch_loss.append(loss.item())
            epoch_f1.append(f1_score(y_true, preds, average=average))

    return np.mean(epoch_f1)


def train_evaluate(
        model: BertCLF,
        training_generator: torch.utils.data.DataLoader,
        valid_generator: torch.utils.data.DataLoader,
        criterion: torch.optim,
        optimizer: torch.nn,
        num_epochs: int,
        average: str,
        config: Dict[str, Dict[str, Any]]
):
    """
    Training and evaluation process
    :param model: architecture you want to fine-tune
    :param training_generator: bert_clf data
    :param valid_generator: evaluation data
    :param criterion: loss from torch losses
    :param optimizer: optimizer from torch optimizers
    :param num_epochs: number of epochs,
    :param config: configuration file with all the parameters needed for training the model,
    :param average: f1-averaging

    :return: fine-tuned model
    """

    stopper = EarlyStopping(
        config=config,
    )
    for i in range(num_epochs):

        print(f"==== Epoch {i+1} out of {num_epochs} ====")
        tr = train(
            model=model,
            iterator=training_generator,
            optimizer=optimizer,
            criterion=criterion,
            average=average
        )

        evl = evaluate(
            model=model,
            iterator=valid_generator,
            criterion=criterion,
            average=average
        )

        stopper(model=model, val_loss=evl)

        if stopper.early_stop:
            print('Early stopping')
            print(f'Train F1: {tr}\nEval F1: {evl}')
            print("\n\n")
            predict_metrics(
                model=model,
                iterator=valid_generator
            )
            return model

        print(f'Train F1: {tr}\nEval F1: {evl}')
        print()

    print()
    predict_metrics(
        model=model,
        iterator=valid_generator
    )
    return model



import torch
from typing import List, Union, Tuple
import numpy as np
from transformers.tokenization_utils_base import PreTrainedTokenizerBase


class Dataset(torch.utils.data.Dataset):

    def __init__(
            self,
            tokenizer: PreTrainedTokenizerBase,
            maxlen: int,
            texts: Union[List[str], np.ndarray],
            targets: Union[List[str], np.ndarray]
    ):
        self.tokenizer = tokenizer
        self.texts = self.texts = [torch.LongTensor(self.tokenizer.encode(
            t,
            truncation=True,
            max_length=maxlen
        )) for t in texts]
        self.texts = torch.nn.utils.rnn.pad_sequence(self.texts, batch_first=True,
                                                     padding_value=self.tokenizer.pad_token_id)

        self.length = len(texts)

        self.target = torch.LongTensor(targets)

    def __len__(self) -> int:
        return self.length

    def __getitem__(self, index: int) -> Tuple[torch.LongTensor, torch.LongTensor]:
        ids = self.texts[index]
        y = self.target[index]

        return ids, y



import numpy as np


class EarlyStopping:
    """Early stops the training if validation loss doesn't improve after a given patience."""

    def __init__(self, config=None):
        """
        Args:
            config (dict): Parameters for training the model
        """
        self.patience = config['training']['patience']
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf
        self.delta = config['training']['delta']
        self.config = config

    def __call__(self, val_loss, model):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            self.val_loss_min = val_loss

        elif score < self.best_score + self.delta:
            self.counter += 1
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.val_loss_min = val_loss
            self.counter = 0



import torch
import torch.nn as nn
from typing import Dict
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
from transformers.modeling_utils import PreTrainedModel


class BertCLF(nn.Module):

    def __init__(
            self,
            pretrained_model: PreTrainedModel,
            tokenizer: PreTrainedTokenizerBase,
            id2label: Dict[int, str],
            dropout: float,
            tiny: bool = True,
            device: torch.device = torch.device('cpu')
    ):
        super().__init__()
        self.tokenizer = tokenizer
        self.pretrained_model = pretrained_model
        self.drop = nn.Dropout(dropout)
        self.act = nn.LogSoftmax(1)
        self.mapper = id2label
        self.device = device

        out_bert = 768
        if tiny:
            out_bert = 312
        self.fc = nn.Linear(out_bert, len(self.mapper))

    def forward(self, texts: torch.Tensor) -> torch.Tensor:
        mask = (texts != self.tokenizer.pad_token_id).long()

        hidden = self.pretrained_model(texts, attention_mask=mask)[0]

        dense_outputs = self.fc(self.drop(hidden[:, 0]))
        outputs = self.act(dense_outputs)

        return outputs

    def predict(self, text: str) -> str:
        inputs = self.tokenizer.encode(text, return_tensors="pt", truncation=True)
        outputs = self(inputs)
        pred = outputs.argmax(1).item()
        pred_text = self.mapper[pred]
        return pred_text



import os
import json
import torch
import torch.optim as optim
import torch.nn as nn
from transformers import AutoModel, AutoTokenizer
from bert_clf.utils import load_config, get_argparse, set_global_seed
from bert_clf.src.training_utils import train_evaluate
from bert_clf.src.BertCLF import BertCLF
from bert_clf.src.preparing_data_utils import prepare_data, prepare_dataset


def train(path_to_config: str):
    """
    path_to_config: path to yaml config file with all the information concerning the training
    """
    config = load_config(path_to_config)

    set_global_seed(seed=config['data']['random_state'])

    os.makedirs(config['training']['output_dir'], exist_ok=True)

    device = torch.device(config['transformer_model']['device'])
    tokenizer = AutoTokenizer.from_pretrained(
        pretrained_model_name_or_path=config['transformer_model']["model"]
    )
    model_bert = AutoModel.from_pretrained(
        pretrained_model_name_or_path=config['transformer_model']["model"]
    ).to(device)

    id2label, train_texts, valid_texts, train_targets, valid_targets = prepare_data(config=config)

    model = BertCLF(
        pretrained_model=model_bert,
        tokenizer=tokenizer,
        id2label=id2label,
        dropout=config['transformer_model']['dropout'],
        tiny=config['transformer_model']['tiny_bert'],
        device=device
    )

    model = model.to(device)

    if config['transformer_model']["path_to_state_dict"]:
        model.load_state_dict(
            torch.load(config['transformer_model']["path_to_state_dict"], map_location=device),
            strict=False
        )

    optimizer = optim.Adam(model.parameters(), lr=float(config['transformer_model']['learning_rate']))
    criterion = nn.NLLLoss()

    training_generator, valid_generator = prepare_dataset(
        tokenizer=tokenizer,
        train_texts=train_texts,
        train_targets=train_targets,
        valid_texts=valid_texts,
        valid_targets=valid_targets,
        config=config
    )

    model = train_evaluate(
        model=model,
        training_generator=training_generator,
        valid_generator=valid_generator,
        criterion=criterion,
        optimizer=optimizer,
        num_epochs=config['training']['num_epochs'],
        average=config['training']['average_f1'],
        config=config
    )

    if config['training']['save_state_dict']:
        torch.save(
            model.state_dict(),
            os.path.join(config["training"]["output_dir"], "model.pth"),
        )

        with open(
                os.path.join(config["training"]["output_dir"], 'label_mapper.json'),
                mode='w',
                encoding='utf-8'
        ) as f:
            json.dump(model.mapper, f, indent=4, ensure_ascii=False)
    else:
        torch.save(
            model,
            os.path.join(config["training"]["output_dir"], "model.pth"),
        )


def main():
    parser = get_argparse()
    args = parser.parse_args()

    train(path_to_config=args.path_to_config)


if __name__ == "__main__":
    main()



import pandas as pd
from pathlib import Path
from bert_clf.src.pandas_dataset.BaseDataset import BaseDataset
import os


class PandasDataset(BaseDataset):
    """
    Dataset class for datasets that have simple structure
    """
    def __init__(self,
                 train_data_path: str,
                 test_data_path: str = None,
                 random_state: int = 42,
                 text_label: str = '',
                 target_label: str = '',
                 test_size: float = 0.3
                 ):

        self.valid_data_types = {
            '.csv': self._read_csv,
            '.tsv': self._read_csv,
            '.xls': self._read_excel,
            '.xlsx': self._read_excel,
            '.json': self._read_json,
            '.jsonl': self._read_json
        }

        super(PandasDataset, self).__init__(
            train_data_path=train_data_path,
            test_data_path=test_data_path,
            random_state=random_state,
            text_label=text_label,
            target_label=target_label,
            test_size=test_size
        )

    def read_data(self, path: Path) -> pd.DataFrame:
        """
        Given the path to the file returns extension of that file

        Example:
        path: "../input/some_data.csv"
        :return: ".csv"
        """
        _, extension = os.path.splitext(path)
        if extension.lower() in self.valid_data_types:
            return self.valid_data_types[extension](path=path, extension=extension)
        else:
            raise ValueError(f"Your data type ({extension}) is not supported, please convert your dataset "
                             f"to one of the following formats {list(self.valid_data_types.keys())}.")

    @staticmethod
    def _read_csv(path: Path, extension: str) -> pd.DataFrame:
        """
        Reads a csv file given its path
        :param path: Path("../../some_file.csv")
        :return: dataframe
        """
        sep = ','
        if extension == '.tsv':
            sep = '\t'

        return pd.read_csv(filepath_or_buffer=path, sep=sep, encoding="utf-8")

    @staticmethod
    def _read_excel(path: Path, extension: str) -> pd.DataFrame:
        """
        Reads a xls or xlsx file given its path
        :param path: Path("../../some_file.xlsx")
        :return: dataframe
        """
        engine = 'openpyxl'
        if extension == '.xls':
            engine = None

        return pd.read_excel(io=path, engine=engine)

    @staticmethod
    def _read_json(path: Path, extension: str) -> pd.DataFrame:
        """
        Reads a json or jsonl file given its path
        :param path: Path("../../some_file.jsonl")
        :return: dataframe
        """
        lines = False
        if extension == '.jsonl':
            lines = True
        return pd.read_json(path_or_buf=path, lines=lines, encoding="utf-8")





import yaml
from argparse import ArgumentParser
import torch
import random
import numpy as np


def load_config(path: str):
    with open(path, "r") as yamlfile:
        data = yaml.safe_load(yamlfile)
        return data


def get_argparse() -> ArgumentParser:
    """Get argument parser.
    Returns:
        ArgumentParser: Argument parser.
    """

    parser = ArgumentParser(prog="bert-clf-train")
    parser.add_argument(
        "--path_to_config",
        type=str,
        required=True,
        help="Path to config",
    )

    return parser


def set_global_seed(seed: int):
    """
    Set global seed for reproducibility.
    """

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True


set_global_seed(42)


from setuptools import setup, find_packages

setup(
    name="bert-for-sequence-classification",
    version='0.0.2',
    author="Tatiana Iazykova",
    author_email="tania_yazykova@bk.ru",
    description='Easy fine-tuning for BERT models',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'bert-clf-train = bert_clf.pipeline:main',
        ],
    },
    install_requires=[
        'transformers==4.2.0',
        'torch>=1.7.1, != 1.10',
        'numpy>=1.19.5',
        'pandas>=1.1.5',
        'scikit-learn>=1.0',
        'pyyaml>=6.0',
        'openpyxl>=3.0.9',
    ],

    keywords=['python', 'bert', 'deep learning', 'nlp'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ]
)




import pandas as pd
from bert_clf.src.pandas_dataset.PandasDataset import PandasDataset
from bert_clf.src.pandas_dataset.SimpleDataFrame import SimpleDataFrame
from typing import Dict, Any, List, Tuple, Optional, Union
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
from bert_clf.src.dataset import Dataset
import torch


def prepare_data(
        config: Dict[str, Dict[str, Any]],
) -> Tuple[Dict[int, str], List[str], List[str], List[int], List[int]]:
    """
    loads data and encodes targets

    :param config: config with  all the necessary information for loading data
    :return: mapper and separated texts and labels
    """

    df = PandasDataset(
        train_data_path=config['data']['train_data_path'],
        test_data_path=config['data']['test_data_path'],
        random_state=config['data']['random_state'],
        text_label=config['data']['text_column'],
        target_label=config['data']['target_column'],
        test_size=config['data']['test_size']
    )

    return get_mapper_and_separated_data(config=config, df=df)


def prepare_data_notebook(
        config: Dict[str, Dict[str, Any]],
        train_df: pd.DataFrame,
        test_df: Optional[pd.DataFrame] = None
) -> Tuple[Dict[int, str], List[str], List[str], List[int], List[int]]:
    """
    loads data and encodes targets

    :param config: config with  all the necessary information for loading data
    :param train_df: bert_clf dataframe
    :param test_df: testing dataframe
    :return: mapper and separated texts and labels
    """

    df = SimpleDataFrame(
        train_data=train_df,
        test_data=test_df,
        random_state=config['data']['random_state'],
        text_label=config['data']['text_column'],
        target_label=config['data']['target_column'],
        test_size=config['data']['test_size']
    )

    return get_mapper_and_separated_data(config=config, df=df)


def get_mapper_and_separated_data(
        config: Dict[str, Dict[str, Any]],
        df: Union[SimpleDataFrame, PandasDataset]
) -> Tuple[Dict[int, str], List[str], List[str], List[int], List[int]]:
    """
    separates texts and targets and encodes the latter

    :param config: config with  all the necessary information for handling the data
    :param df: class object that has train and valid data

    :return: mapper and separated texts and labels
    """

    id2label = {i: l for i, l in enumerate(df.train[config['data']['target_column']].unique())}
    label2id = {l: i for i, l in id2label.items()}
    train_texts = df.train[config['data']['text_column']].to_list()
    valid_texts = df.test[config['data']['text_column']].to_list()
    train_targets = df.train[config['data']['target_column']].map(label2id).to_list()
    valid_targets = df.test[config['data']['target_column']].map(label2id).to_list()

    return id2label, train_texts, valid_texts, train_targets, valid_targets


def prepare_dataset(
        tokenizer: PreTrainedTokenizerBase,
        train_texts: List[str],
        train_targets: List[int],
        valid_texts: List[str],
        valid_targets: List[int],
        config: Dict[str, Dict[str, Any]]
) -> Tuple[torch.utils.data.DataLoader, torch.utils.data.DataLoader]:
    """

    :param tokenizer: tokenizer from transformers
    :param train_texts: list of texts selected for bert_clf
    :param train_targets: list of targets for texts selected for bert_clf
    :param valid_texts: list of texts selected for evaluation
    :param valid_targets: list of targets for texts selected for evaluation
    :param config:  config with all the necessary information to set up a model

    :return: Dataloader  objects for bert_clf and evaluatuon
    """

    training_set = Dataset(
        tokenizer=tokenizer,
        texts=train_texts,
        targets=train_targets,
        maxlen=config['transformer_model']['maxlen']
    )

    training_generator = torch.utils.data.DataLoader(
        training_set,
        batch_size=config['transformer_model']['batch_size'],
        shuffle=config['transformer_model']['shuffle']
    )

    valid_set = Dataset(
        tokenizer=tokenizer,
        texts=valid_texts,
        targets=valid_targets,
        maxlen=config['transformer_model']['maxlen']
    )

    valid_generator = torch.utils.data.DataLoader(
        valid_set,
        batch_size=config['transformer_model']['batch_size'],
        shuffle=config['transformer_model']['shuffle']
    )

    return training_generator, valid_generator
